# generate variadic calls to test the
# abi implementation

from random import seed, randint, uniform
from struct import unpack

I, D = 'd', 'g'

formats = [
	# list of formats to test
	[I],
	[D],
	[I,D],
	[D,D],
	[I,I,I,I],
	[D,D,D,D],
	[I,D,I,D],
	[D,D,I,I],
	[I,I,D,D],
	[],
]

generate = [
	# numbers of fixed integer and
	# floating point arguments to
	# test
	(0, 0), (1, 0), (0, 1), (4, 0),
	(0, 6), (5, 7), (10, 10), (9, 0),
]

def mkargs(nargs, type, name):
	args = map(
		lambda n: ''.join([type, name, str(n), ', ']),
		range(nargs)
	)
	return ''.join(args)

def mkfstr(fmt):
	fstr = map(
		lambda x: {I: '%d ', D: '%g '}[x],
		fmt
	)
	return '"' + ''.join(fstr) + '\\n"'

def randargs(fmt):
	ra = {
		I: lambda: '{}'.format(randint(-10, 10)),
		D: lambda: '{0:.4g}'.format(uniform(-10, 10))
	}
	return list(map(lambda x: ra[x](), fmt))

def genssa(qbeprint, qbecall):
	funcs = [('qbeprint', qbeprint), ('qbecall', qbecall)]
	for fnum, (nia, nfa) in enumerate(generate):
		params = "{}{}l %fmt, ...".format(
			mkargs(nia, 'w ', '%argw'),
			mkargs(nfa, 'd ', '%argd')
		)
		for name, code in funcs:
			print('export function ${}{}({}) {}'
				.format(name, fnum, params, code)
			)

def gendriver():
	print('# >>> driver')
	print('# #include <stdarg.h>')
	print('# #include <stdio.h>')

	for fnum, (nia, nfa) in enumerate(generate):
		params = "{}{}char *, ...".format(
			mkargs(nia, 'int ', 'argw'),
			mkargs(nfa, 'double ', 'argd')
		)
		for name in ['qbeprint', 'qbecall']:
			print('# extern void {}{}({});'
				.format(name, fnum, params)
			)

	output = ''
	print('# int print(char *fmt, va_list *ap) {')
	print('#    return vprintf(fmt, *ap);');
	print('# }')
	print('# int main() {')

	for fnum, (nia, nfa) in enumerate(generate):
		info = '# ({} int, {} double)'.format(nia, nfa)
		print('# 	puts("{}");'.format(info))
		output += '# {}\n'.format(info)
		for fmt in formats:
			ra = randargs(fmt)
			vaargs = ', '.join(ra)
			expect = ' '.join(ra)
			if fmt:
				vaargs = ', ' + vaargs
				expect = expect + ' '
			args = ''.join(
				['0, '] * (nia+nfa) +
				[mkfstr(fmt), vaargs]
			)
			for name in ['qbeprint', 'qbecall']:
				print('# 	{}{}({});'
					.format(name, fnum, args)
				)
				output += '# {}\n'.format(expect)

	print('# }')
	print('# <<<')

	print('\n# >>> output\n' + output + '# <<<')


qbeprint="""{{
@start
	%fmtdbl =l alloc4 4
	%fmtint =l alloc4 4
	%emptys =l alloc4 4
	storew {}, %fmtint
	storew {}, %fmtdbl
	storew 0, %emptys
	%vp =l alloc8 32
	%fmt1 =l add 1, %fmt
	vastart %vp
@loop
	%p =l phi @start %fmt1, @casef %p1, @cased %p1
	%c =w loadsb %p
	%p1 =l add 3, %p
	jnz %c, @loop1, @end
@loop1
	%isg =w ceqw %c, 103
	jnz %isg, @casef, @cased
@casef
	%dbl =d vaarg %vp
	%r =w call $printf(l %fmtdbl, ..., d %dbl)
	jmp @loop
@cased
	%int =w vaarg %vp
	%r =w call $printf(l %fmtint, ..., w %int)
	jmp @loop
@end
	%r =w call $puts(l %emptys)
	ret
}}
""".format(
	unpack("i", b'%d \x00')[0],
	unpack("i", b'%g \x00')[0]
)

qbecall="""{
@start
	%vp =l alloc8 32
	vastart %vp
	%r =w call $print(l %fmt, l %vp)
	ret
}
"""


if __name__ == "__main__":
	seed(42)
	genssa(qbeprint, qbecall)
	gendriver()



#!/usr/bin/python3

# support script to create
# the abi8.ssa test

def ctype(arg):
    if arg[0] == 'p': return ctype(arg[1:])
    if arg[0] == ':': return 'S' + arg[1:]
    return {'w':'int', 'l':'long',
            's':'float', 'd':'double'}[arg]

def cparam(iarg):
    return ctype(iarg[1]) + ' p' + str(iarg[0])

def gencfn(id, args):
    out = '# extern void qfn' + id + '('
    out += ', '.join(map(ctype, args)) + ');\n'
    out += '# void cfn' + id + '('
    out += ', '.join(map(cparam, enumerate(args)))
    out += ') {\n'
    out += '# \tprintf("qbe->c(%d)", ' + id + ');\n'
    out += '# \t'
    for (i, arg) in enumerate(args):
        if arg[0] != 'p': continue
        ty = arg[1:]
        if ty[0] == ':':
            out += 'p' + ty[1:] + '(&'
        else:
            out += 'p' + ty + '('
        out += 'p' + str(i) + '); '
    out += 'puts("");\n'
    out += '# \tqfn' + id +  '('
    out += ', '.join('p'+str(i) for i in range(len(args)))
    out += ');\n'
    out += '# }\n'
    return out

def qparam(iarg):
    ty = iarg[1][1:] if iarg[1][0] == 'p' else iarg[1]
    return ty + ' %p' + str(iarg[0])

def genqfn(id, args):
    out = 'export\nfunction $qfn' + id + '('
    out += ', '.join(map(qparam, enumerate(args)))
    out += ') {\n'
    out += '@start\n'
    out += '\t%r0 =w call $printf(l $ctoqbestr, w ' + id + ')\n'
    for (i, arg) in enumerate(args):
        if arg[0] != 'p': continue
        ty = arg[1:]
        if ty[0] == ':':
            out += '\tcall $p' + ty[1:]
            out += '(l %p' + str(i) + ')\n'
        else:
            out += '\tcall $p' + ty
            out += '(' + ty + ' %p' + str(i) + ')\n'
    out += '\t%r1 =w call $puts(l $emptystr)\n'
    out += '\tret\n'
    out += '}\n'
    return out

def carg(iarg):
    i, arg = iarg
    print = arg[0] == 'p'
    ty = arg if not print else arg[1:]
    if ty[0] == ':':
        if print:
            return ty + ' $' + ty[1:]
        else:
            return ty + ' $z' + ty[1:]
    if not print:
        return ty + ' 0'
    if ty == 'w' or ty == 'l':
        return ty + ' ' + str(i+1)
    if ty == 's' or ty == 'd':
        flt = str(i+1) + '.' + str(i+1)
        return ty + ' ' + ty + '_' + flt

def genmaincall(id, args):
    out = '\tcall $cfn' + id + '('
    out += ', '.join(map(carg, enumerate(args)))
    out += ')\n'
    return out

def gen(tvec):
    for i, t in enumerate(tvec):
        print(genqfn(str(i), t), end='')
    print('')
    for i, t in enumerate(tvec):
        print(genmaincall(str(i), t), end='')
    print('')
    for i, t in enumerate(tvec):
        print(gencfn(str(i), t), end='')

TVEC = [
   ['s']*8 + ['ps'],
   ['pw', 'ps', 'p:fi1'],
   ['pw', 'p:fi2', 'ps'],
   ['pw', 'ps', 'p:fi3'],
   ['p:ss'],
   ['d']*7 + ['p:ss', 'ps', 'pl'],
   ['p:lb'],
   ['w']*7 + ['p:lb'],
   ['w']*8 + ['p:lb'],
   [ 'p:big' ],
   ['w']*8 + ['p:big', 'ps', 'pl'],
]

if __name__ == '__main__':
    gen(TVEC)
